
from sleekxmpp import ClientXMPP
from sleekxmpp.thirdparty import socks

from orgassist import templates
from . import Message
from . import log

class XmppBot:
    """
    Interfaces with Jabber and identifies bosses by JID and resource.
    """
    def __init__(self, connect_cfg):
        "Initialize XMPP bot"
        # (Sender JID, [local resource]) -> callback
        self.dispatch_map = {}

        self.jid = connect_cfg.jid
        self.connect(connect_cfg)

    def connect(self, config):
        "Connect to XMPP server"
        socks_cfg = config.get('socks_proxy', required=False)
        if socks_cfg is not None:
            # TODO: Doesn't work yet. Doesn't implement a makelist
            socks.setdefaultproxy(socks.PROXY_TYPE_SOCKS5,
                                  socks_cfg.get('host', assert_type=str),
                                  socks_cfg.get('port', assert_type=int),
                                  rdns=True)
            socket = socks.socksocket
        else:
            socket = None

        self.client = ClientXMPP(self.jid, config.password, socket=socket)
        proxy_cfg = config.get('http_proxy', required=False)
        if proxy_cfg is not None:
            self.client.use_proxy = True
            self.client.proxy_config = {
                'host': proxy_cfg.host,
                'port': proxy_cfg.get('port', assert_type=int),
                'username': proxy_cfg.get('username', required=False),
                'password': proxy_cfg.get('password', required=False),
            }
            print("Configured proxy", self.client.proxy_config)

        use_tls = config.get('tls', default=True)

        # Events
        self.client.add_event_handler("session_start", self._session_start)
        self.client.add_event_handler("message", self.message_dispatch)

        log.info("Initializing connection to XMPP")
        ip = config.get('ip', required=False)
        if ip is not None:
            print("port", config.port)
            port = config.get('port', default=5222, assert_type=int)
            self.client.connect((ip, port), use_tls=use_tls)
        else:
            self.client.connect(use_tls=use_tls)

    def _session_start(self, event):
        log.info('Starting XMPP session')
        self.client.send_presence()

    def add_dispatch(self, sender_jid, resource, callback):
        """
        Register callback to handle incoming messages from sender_jid
        directed to a given resource (can be None to mean "any").
        """
        key = (sender_jid, resource)
        if key in self.dispatch_map:
            raise Exception("Sender JID duplicated: %s" % sender_jid)
        self.dispatch_map[key] = callback

    def message_dispatch(self, msg):
        "Dispatch incoming message depending on the sender"
        if msg['type'] not in ('chat', 'normal'):
            log.warning('Unknown message type: %r %r', msg, msg['type'])
            return

        to_jid = msg.get_to()
        from_jid = msg.get_from()
        resource = to_jid.resource

        def respond(response):
            "Closure to simplify responding"
            self.send_message(from_jid, response)

        log.debug("Got message:")
        log.debug("From: %s, to: %s", from_jid, to_jid)
        log.debug("Body: %r", msg)

        # Dispatch direct (to resource) or generic (to JID)
        callback = self.dispatch_map.get((from_jid.bare, resource), None)
        if callback is None:
            callback = self.dispatch_map.get((from_jid.bare, None), None)

        # If unknown - ignore
        if callback is None:
            respond(templates.get('DONT_KNOW'))
            return

        # Construct a Message using bot-generic API
        message = Message(msg['body'], from_jid.full, respond)

        callback(message)

    def send_message(self, jid, message):
        "Send a message"
        self.client.send_message(jid, message)

    def close(self):
        "Disconnect / close threads"
        self.client.abort()



import unittest
import random
import datetime as dt

from orgassist.calendar import EventDate, Event, DateType
from orgassist.calendar import EventState
from orgassist.calendar import Calendar

class TestEvent(unittest.TestCase):
    """
    Test Event and EventDates helpers.
    """

    @staticmethod
    def day_starts():
        """
        Early day date.

        Test would fail without an early "now" date because appointments past
        midnight get less important than stuff scheduled for "today" (23:59).

        This relavancy system might work bad for people working graveyard
        shifts.
        """
        return dt.datetime.now().replace(hour=random.randint(6, 10),
                                         minute=random.randint(0, 59))

    def create_dates(self):
        "Shared dates. A bit of ugly code to have low number of repetitions"
        # pylint: disable=attribute-defined-outside-init,too-few-public-methods,too-many-instance-attributes
        class Dates:
            "Convenient date holder"
        dates = Dates()

        dates.now = self.day_starts()
        now = dates.now

        # Test date creation
        dates.appointment_in_hour = EventDate(now + dt.timedelta(hours=1),
                                              DateType.TIMESTAMP)

        dates.deadline_in_hour = EventDate(now + dt.timedelta(hours=1),
                                           DateType.DEADLINE)

        dates.appointment_in_3hours = EventDate(now + dt.timedelta(hours=3),
                                                DateType.TIMESTAMP)

        dates.appointment_later = EventDate(now + dt.timedelta(days=2),
                                            DateType.TIMESTAMP)

        dates.old_appointment = EventDate(now - dt.timedelta(days=2),
                                          DateType.TIMESTAMP)

        dates.task_today = EventDate(now.date(),
                                     DateType.SCHEDULED)

        dates.deadline_yesterday = EventDate((now - dt.timedelta(days=1)).date(),
                                             DateType.SCHEDULED)

        dates.start_date = now + dt.timedelta(minutes=10)
        dates.end_date = now + dt.timedelta(days=3)
        dates.ranged_date = EventDate((dates.start_date,
                                       dates.end_date),
                                      DateType.RANGE)
        return dates

    def test_dates(self):
        "Test event date creation"
        dates = self.create_dates()
        now = dates.now

        # Asserts on the API
        self.assertEqual(dates.ranged_date.date, dates.start_date)
        self.assertEqual(dates.ranged_date.date_end, dates.end_date)
        self.assertEqual(dates.ranged_date.date_type, DateType.RANGE)
        self.assertTrue(dates.ranged_date.appointment)

        self.assertTrue(dates.appointment_in_3hours.appointment)
        self.assertFalse(dates.task_today.appointment)

        # Check relevancy algorithm
        self.assertTrue(dates.appointment_in_hour.is_more_relevant(dates.appointment_in_3hours,
                                                                   relative_to=now))
        self.assertTrue(dates.appointment_in_hour.is_more_relevant(dates.appointment_later,
                                                                   relative_to=now))
        self.assertTrue(dates.appointment_in_hour.is_more_relevant(dates.task_today,
                                                                   relative_to=now))
        self.assertFalse(dates.task_today.is_more_relevant(dates.appointment_in_hour,
                                                           relative_to=now))
        self.assertTrue(dates.task_today.is_more_relevant(dates.old_appointment,
                                                          relative_to=now))
        self.assertTrue(dates.appointment_later.is_more_relevant(dates.old_appointment,
                                                                 relative_to=now))

        # Ranged starts earlier
        self.assertFalse(dates.appointment_in_3hours.is_more_relevant(dates.ranged_date,
                                                                      relative_to=now))

        # Mixed thoughts on that. ;-)
        self.assertTrue(dates.task_today.is_more_relevant(dates.deadline_yesterday,
                                                          relative_to=now))
        self.assertTrue(dates.deadline_in_hour.is_more_relevant(dates.appointment_in_hour,
                                                                relative_to=now))

    def test_events(self):
        "Test event behaviour"

        headline = "This is an event headline!"
        event = Event(headline, state="TODO")

        # Asserts on API
        self.assertEqual(event.headline, headline)
        self.assertEqual(event.body, "")
        self.assertIsNone(event.relevant_date)

        # Schedule the event for today.
        now = self.day_starts()
        sched_date = EventDate(now.date(), DateType.SCHEDULED)
        event.add_date(sched_date, relative_to=now)

        self.assertTrue(DateType.SCHEDULED in event.date_types)
        self.assertEqual(event.relevant_date, sched_date)

        # But when given another date:
        app_date = EventDate(now + dt.timedelta(hours=2),
                             DateType.TIMESTAMP)
        event.add_date(app_date, relative_to=now)

        self.assertTrue(DateType.TIMESTAMP in event.date_types)
        self.assertEqual(event.relevant_date, app_date)

        # Order should change anything:
        state = EventState("TODO")
        event = Event(headline, state=state)
        event.add_date(app_date, relative_to=now)
        event.add_date(sched_date, relative_to=now)
        self.assertTrue(DateType.TIMESTAMP in event.date_types)
        self.assertEqual(event.relevant_date, app_date)

    def test_tags(self):
        "Test tags on event"

        headline = "This is an event headline!"
        state = EventState("TODO")
        self.assertTrue(state.is_open)
        event = Event(headline, state=state)
        event.add_tags(["TEST", "PRIVATE"])
        self.assertIn('TEST', event.tags)

    def test_calendar(self):
        "Test calendar behaviour"
        dates = self.create_dates()
        events = []
        for name, value in dates.__dict__.items():
            if isinstance(value, EventDate):
                event = Event(name)
                event.add_date(value)
                events.append(event)

        horizon_incoming = dates.now.replace(hour=23, minute=59, second=59)
        horizon_unfinished = dates.now - dt.timedelta(days=2)

        agenda_template = """
        {% for task in unfinished %}
         Unfinished:{{ task }}
        {% endfor %}
        {% for task in incoming %}
         Incoming:{{ task }}
        {% endfor %}
        """

        calendar = Calendar(agenda_content=agenda_template)
        calendar.add_events(events, internal_tag='org')

        unfinished = calendar.get_unfinished(horizon_unfinished,
                                             list_unfinished_appointments=True,
                                             relative_to=dates.now)
        scheduled = calendar.get_scheduled(horizon_incoming, relative_to=dates.now)
        agenda = calendar.get_agenda(horizon_incoming=horizon_incoming,
                                     horizon_unfinished=horizon_unfinished,
                                     list_unfinished_appointments=True)

        print("AGENDA:", type(agenda), agenda, "END")
        #self.assertGreaterEqual(len(unfinished), 2)
        #self.assertGreaterEqual(len(scheduled), 0)
        #self.assertGreaterEqual(len(agenda.split('\n')), 5)



import random

_templates = {
    'DONT_KNOW': "I don't know you. Shush!",
    'DONT_UNDERSTAND': [
        "Excuse me?",
        "Sorry?",
        "Can you repeat?",
    ],
}

def get(key):
    value = _templates.get(key)
    if isinstance(value, list):
        return random.choice(value)
    return value



"""
The Orgnode module consists of the Orgnode class for representing a
headline and associated text from an org-mode file, and routines for
constructing data structures of these classes.

LICENSE:

Copyright (c) 2012 Albin Stjerna, Takafumi Arakaki, and Charles Cave

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

https://github.com/albins/orgnode.git
"""

import re
import datetime
import codecs

def get_datetime(year, month, day, hour=None, minute=None, second=None):
    if "" in (year, month, day):
        raise ValueError("First three arguments must not contain empty str")
    if None in (year, month, day):
        raise ValueError("First three arguments must not contain None")

    ymdhms = []
    for a in [year, month, day, hour, minute, second]:
        if a != None and a != "":
            ymdhms.append(int(a))

    if len(ymdhms) > 3:
        return datetime.datetime(*ymdhms)
    return datetime.date(*ymdhms)

def _re_compile_date():
    """
    >>> re_date = _re_compile_date()
    >>> re_date.match('')
    >>> m = re_date.match('<2010-06-21 Mon>')
    >>> m.group()
    '<2010-06-21 Mon>'
    >>> m.group(1)
    >>> m.group(16)
    '<2010-06-21 Mon>'
    >>> m = re_date.match('<2010-06-21 Mon 12:00>--<2010-06-21 Mon 12:00>')
    >>> m.group()
    '<2010-06-21 Mon 12:00>--<2010-06-21 Mon 12:00>'
    >>> m.group(1)
    '<2010-06-21 Mon 12:00>--<2010-06-21 Mon 12:00>'
    >>> m.group(16)
    """
    date_pattern = r"<(\d+)\-(\d+)\-(\d+)([^>\d]*)((\d+)\:(\d+))?>"
    re_date = re.compile('(%(dtp)s--%(dtp)s)|(%(dtp)s)'
                         % dict(dtp=date_pattern))
    return re_date

_RE_DATE = _re_compile_date()
def find_daterangelist(string):
    datelist = []
    rangelist = []
    for dm in _RE_DATE.findall(string):
        if dm[0]:
            d1 = get_datetime(dm[1], dm[2], dm[3], dm[6], dm[7])
            d2 = get_datetime(dm[8], dm[9], dm[10], dm[13], dm[14])
            rangelist.append((d1, d2))
        else:
            dt = get_datetime(dm[16], dm[17], dm[18], dm[21], dm[22])
            datelist.append(dt)
    return (datelist, rangelist)

_RE_SCHEDULED = re.compile(
    r'SCHEDULED:\s+<(\d+)\-(\d+)\-(\d+)[^>\d]*((\d+)\:(\d+))?>')
def find_scheduled(line):
    """
    Find SCHEDULED from given string.
    Return datetime object if found else None.
    """
    sd_re = _RE_SCHEDULED.search(line)
    if sd_re:
        if sd_re.group(4) == None:
            sched_date = datetime.date(int(sd_re.group(1)),
                                       int(sd_re.group(2)),
                                       int(sd_re.group(3)))
        else:
            sched_date = datetime.datetime(int(sd_re.group(1)),
                                           int(sd_re.group(2)),
                                           int(sd_re.group(3)),
                                           int(sd_re.group(5)),
                                           int(sd_re.group(6)))
    else:
        sched_date = None
    return sched_date

_RE_DEADLINE = re.compile(r'DEADLINE:\s+<(\d+)\-(\d+)\-(\d+)[^>\d]*((\d+)\:(\d+))?>')

def find_deadline(line):
    """
    Find DEADLINE from given string.
    Return datetime object if found else None.
    """
    dd_re = _RE_DEADLINE.search(line)
    if dd_re:
        if dd_re.group(4) == None:
            deadline_date = datetime.date(int(dd_re.group(1)),
                                          int(dd_re.group(2)),
                                          int(dd_re.group(3)))
        else:
            deadline_date = datetime.datetime(int(dd_re.group(1)),
                                              int(dd_re.group(2)),
                                              int(dd_re.group(3)),
                                              int(dd_re.group(5)),
                                              int(dd_re.group(6)))
    else:
        deadline_date = None
    return deadline_date

_RE_TAGSRCH = re.compile(r'(.*?)\s*:(.*?):(.*?)$')
def find_tags_and_heading(heading):
    """
    Get first tag, all tags, and heading without tags.
    This is helper function of makelist.
    """
    tag1 = ""
    alltags = set() # set of all tags in headline
    tagsrch = _RE_TAGSRCH.search(heading)
    if tagsrch:
        heading = tagsrch.group(1)
        tag1 = tagsrch.group(2)
        alltags.add(tag1)
        tag2 = tagsrch.group(3)
        if tag2:
            alltags |= set(tag2.split(':')) - set([''])
    return (tag1, alltags, heading)

_RE_PROP_SRCH = re.compile(r'^\s*:(.*?):\s*(.*?)\s*$')
def find_property(line):
    """
    Find property from given string.
    Return (key, value)-pair if found else (None, None).
    """
    prop_key = None
    prop_val = None
    prop_srch = _RE_PROP_SRCH.search(line)
    if prop_srch:
        prop_key = prop_srch.group(1)
        prop_val = prop_srch.group(2)
        if prop_key == 'Effort':
            if ":" in prop_val:
                (h, m) = prop_val.split(":", 2)
            else:
                h, m = prop_val, "0"
            if h.isdigit() and m.isdigit():
                prop_val = int(h)*60 + int(m)
    return (prop_key, prop_val)


_RE_CLOSED = re.compile(r'CLOSED:\s+\[(\d+)\-(\d+)\-(\d+)[^\]\d]*((\d+)\:(\d+))?\]')
def find_closed(line):
    """
    Find CLOSED from given string.
    Return datetime object if found else None.
    """
    cl_re = _RE_CLOSED.search(line)
    if cl_re:
        if cl_re.group(4) == None:
            closed_date = datetime.date(int(cl_re.group(1)),
                                        int(cl_re.group(2)),
                                        int(cl_re.group(3)))
        else:
            closed_date = datetime.datetime(int(cl_re.group(1)),
                                            int(cl_re.group(2)),
                                            int(cl_re.group(3)),
                                            int(cl_re.group(5)),
                                            int(cl_re.group(6)))
    else:
        closed_date = None
    return closed_date


_RE_CLOCK = re.compile(
    r'CLOCK:\s+'
    r'\[(\d+)\-(\d+)\-(\d+)[^\]\d]*(\d+)\:(\d+)\]--'
    r'\[(\d+)\-(\d+)\-(\d+)[^\]\d]*(\d+)\:(\d+)\]\s+=>\s+(\d+)\:(\d+)'
    )

def find_clock(line):
    """
    Find CLOCK from given string.
    Return three tuple (start, stop, length) which is datetime object
    of start time, datetime object of stop time and length in minute.
    """
    match = _RE_CLOCK.search(line)
    if match is None:
        return None
    groups = [int(d) for d in match.groups()]
    ymdhm1 = groups[:5]
    ymdhm2 = groups[5:10]
    hm3 = groups[10:]
    return (
        datetime.datetime(*ymdhm1),
        datetime.datetime(*ymdhm2),
        hm3[0]*60 + hm3[1],
        )

_RE_HEADING = re.compile(r'^(\*+)\s(.*?)\s*$')
_RE_TODO_KWDS = re.compile(r' ([A-Z][A-Z0-9]+)\(?')
_RE_TODO_SRCH = re.compile(r'^\s*([A-Z][A-Z0-9]+)\s(.*?)$')
_RE_PRTY_SRCH = re.compile(r'^\[\#(A|B|C)\] (.*?)$')

def makelist(filename, todo_default=['TODO', 'DONE']):
    """
    Read an org-mode file and return a list of Orgnode objects
    created from this file.
    """
    ctr = 0

    if isinstance(filename, str):
        f = codecs.open(filename, 'r', 'utf8')
    else:
        f = filename

    todos = set(todo_default) # populated from #+SEQ_TODO line
    level = ''
    heading = ""
    bodytext = ""
    tag1 = ""      # The first tag enclosed in ::
    alltags = set([]) # set of all tags in headline
    sched_date = ''
    deadline_date = ''
    closed_date = ''
    clocklist = []
    datelist = []
    rangelist = []
    nodelist = []
    propdict = dict()

    for line in f:
        ctr += 1
        hdng = _RE_HEADING.search(line)

        if hdng:
            if heading:  # we are processing a heading line
                this_node = Orgnode(level, heading, bodytext, tag1, alltags)
                if sched_date:
                    this_node.set_scheduled(sched_date)
                    sched_date = ""
                if deadline_date:
                    this_node.set_deadline(deadline_date)
                    deadline_date = ''
                if closed_date:
                    this_node.set_closed(closed_date)
                    closed_date = ''
                if clocklist:
                    this_node.set_clock(clocklist)
                    clocklist = []
                if datelist:
                    this_node.set_datelist(datelist)
                    datelist = []
                if rangelist:
                    this_node.set_rangelist(rangelist)
                    rangelist = []
                this_node.set_properties(propdict)
                nodelist.append(this_node)
                propdict = dict()
            level = hdng.group(1)
            heading = hdng.group(2)
            bodytext = ""
            (tag1, alltags, heading) = find_tags_and_heading(heading)
        else:      # we are processing a non-heading line
            if line.startswith('#+SEQ_TODO'):
                todos |= set(_RE_TODO_KWDS.findall(line))
                continue
            if line.find(':PROPERTIES:') >= 0:
                continue
            if line.find(':END:') >= 0:
                continue
            (prop_key, prop_val) = find_property(line)
            if prop_key:
                propdict[prop_key] = prop_val
                continue
            _sched_date = find_scheduled(line)
            _deadline_date = find_deadline(line)
            _closed_date = find_closed(line)
            sched_date = _sched_date or sched_date
            deadline_date = _deadline_date or deadline_date
            closed_date = closed_date or _closed_date
            if not _sched_date and not _deadline_date:
                (dl, rl) = find_daterangelist(line)
                datelist += dl
                rangelist += rl
            clock = find_clock(line)
            if clock:
                clocklist.append(clock)
            if not (line.startswith('#') or _sched_date or _deadline_date
                    or clock or _closed_date):
                bodytext = bodytext + line

    # write out last node
    this_node = Orgnode(level, heading, bodytext, tag1, alltags)
    this_node.set_properties(propdict)
    if sched_date:
        this_node.set_scheduled(sched_date)
    if deadline_date:
        this_node.set_deadline(deadline_date)
    if closed_date:
        this_node.set_closed(closed_date)
        closed_date = ''
    if clocklist:
        this_node.set_clock(clocklist)
        clocklist = []
    if datelist:
        this_node.set_datelist(datelist)
        datelist = []
    if rangelist:
        this_node.set_rangelist(rangelist)
        rangelist = []
    nodelist.append(this_node)

    # using the list of TODO keywords found in the file
    # process the headings searching for TODO keywords
    for n in nodelist:
        h = n.headline

        todo_search = _RE_TODO_SRCH.search(h)

        if todo_search:
            if todo_search.group(1) in todos:
                n.set_heading(todo_search.group(2))
                n.set_todo(todo_search.group(1))
        priority_search = _RE_PRTY_SRCH.search(n.headline)
        if priority_search:
            n.set_priority(priority_search.group(1))
            n.set_heading(priority_search.group(2))

    # set parent of nodes
    ancestors = [None]
    n1 = nodelist[0]
    l1 = n1.level
    for n2 in nodelist:
        # n1, l1: previous node and its level
        # n2, l2: this node and its level
        l2 = n2.level
        if l1 < l2:
            ancestors.append(n1)
        else:
            while len(ancestors) > l2:
                ancestors.pop()
        if ancestors:
            n2.set_parent(ancestors[-1])
        n1 = n2
        l1 = l2

    return nodelist

######################
class Orgnode:
    """
    Orgnode class represents a headline, tags and text associated
    with the headline.
    """
    def __init__(self, level, headline, body, tag, alltags):
        """
        Create an Orgnode object given the parameters of level (as the
        raw asterisks), headline text (including the TODO tag), and
        first tag. The makelist routine postprocesses the list to
        identify TODO tags and updates headline and todo fields.
        """
        self.level = len(level)
        self.headline = headline
        self.body = body
        self.tag = tag            # The first tag in the list
        self.tags = set(alltags)  # All tags in the headline
        self.todo = ""
        self.priority = ""            # empty of A, B or C
        self.scheduled = ""       # Scheduled date
        self.deadline = ""        # Deadline date
        self.clock = []
        self.closed = ""
        self.properties = dict()
        self.datelist = []
        self.rangelist = []
        self.parent = None

        # Look for priority in headline and transfer to priority field

    def set_heading(self, newhdng):
        """
        Change the heading to the supplied string
        """
        self.headline = newhdng

    def set_priority(self, newprty):
        """
        Change the value of the priority of this headline.
        Values values are '', 'A', 'B', 'C'
        """
        self.priority = newprty

    def get_tags(self, inher=False):
        """
        Returns a list of all tags
        For example, :HOME:COMPUTER: would return ['HOME', 'COMPUTER']
        If `inher` is True, then all tags from ancestors is included.
        """
        if inher and self.parent:
            return self.tags | set(self.parent.get_tags(True))
        else:
            return self.tags

    def has_tag(self, srch):
        """
        Returns True if the supplied tag is present in this headline
        For example, hasTag('COMPUTER') on headling containing
        :HOME:COMPUTER: would return True.
        """
        return srch in self.tags

    def set_tag(self, newtag):
        """
        Change the value of the first tag to the supplied string
        """
        self.tag = newtag

    def set_tags(self, taglist):
        """
        Store all the tags found in the headline. The first tag will
        also be stored as if the setTag method was called.
        """
        self.tags |= set(taglist)

    def set_todo(self, value):
        """
        Set the value of the TODO tag to the supplied string
        """
        self.todo = value

    def set_properties(self, dictval):
        """
        Sets all properties using the supplied dictionary of
        name/value pairs
        """
        self.properties = dictval

    def get_property(self, keyval):
        """
        Returns the value of the requested property or null if the
        property does not exist.
        """
        return self.properties.get(keyval, "")

    def set_scheduled(self, dateval):
        """
        Set the scheduled date using the supplied date object
        """
        self.scheduled = dateval

    def set_deadline(self, dateval):
        """
        Set the deadline (due) date using the supplied date object
        """
        self.deadline = dateval

    def set_datelist(self, datelist):
        """
        Set the list of date using list of the supplied date object
        """
        self.datelist = datelist[:]

    def get_datelist(self):
        """
        Return the list of all date as date object
        """
        return self.datelist[:]

    def set_rangelist(self, rangelist):
        """
        Set the list of date range using list of the supplied timedelta object
        """
        self.rangelist = rangelist[:]

    def get_rangelist(self):
        """
        Return the list of all date as date object
        """
        return self.rangelist[:]

    def has_date(self):
        "True if node has date"
        return (bool(self.scheduled) or
                bool(self.deadline) or
                bool(self.datelist) or
                bool(self.rangelist))

    def set_closed(self, closed):
        """
        Set the closed date using the supplied date object
        """
        self.closed = closed

    def set_clock(self, clock):
        """
        Set the list of clocked time using list of three tuple
        (start, stop, length) which is datetime object of start time,
        datetime object of stop time and length in minute.
        """
        self.clock = clock[:]

    def set_parent(self, parent):
        """
        Set parent node
        """
        self.parent = parent

    def get_root(self):
        """
        Return root node

        What is root node?::

          * I am node A, a root node       <- Root node!
          ** I am a child of node A
          ** I am a child of node A too
          * I am also a root node            <- Root node!

        """
        child = self
        while True:
            parent = child.parent
            if parent is None:
                return child
            child = parent

    def __repr__(self):
        """
        Print the level, heading text and tag of a node and the body
        text as used to construct the node.
        """
        # This method is not completed yet.
        txt = '*' * self.level
        txt = txt + ' ' + self.todo + ' '
        if self.priority:
            txt = txt +  '[#' + self.priority + '] '
        txt = txt + str(self.headline)
        txt = "%-60s " % txt     # hack - tags will start in column 62
        closecolon = ''
        for t in sorted(self.tags):
            txt = txt + ':' + t
            closecolon = ':'
        txt = txt + closecolon
# Need to output Scheduled Date, Deadline Date, property tags The
# following will output the text used to construct the object
        txt = txt + "\n" + repr(self.body)

        return txt



import random
import io
import datetime as dt

import unittest

import jinja2

from orgassist.calendar import DateType
from . import orgnode
from . import helpers

# Example Org file for testing
# pre-generated to have a todays dates.
ORG_TMPL = """
* PROJECT Aggregator
** TODO This is open task   :OPEN_TASK:
   SCHEDULED: <{{ today }}>
** TODO Past task           :TAG1:
   SCHEDULED: <{{ yesterday }}>
** DONE Already done        :DEADLINE:
   DEADLINE: <{{ today }}>
** [#B] Appointment         :APP:
   <{{ accurate }}>
** Whole day event          :WHOLE_DAY:
   <{{ today }}>
** Inactive date            :INACTIVE:
   [{{ accurate }}]
** [#A] Ranged              :RANGE:
   <{{ yesterday }}>--<{{ today }}>
"""
DAYTIME = '%Y-%m-%d %a %H:%M'
DAY = '%Y-%m-%d %a'

# For testing org helpers
ORG_CONFIG = {
    'files': [],

    'files_re': None,
    'base': None,

    'todos_open': ['TODO'],
    'todos_closed': ['DONE', 'CANCELLED'],

    # How grouping entry is marked - which groups TODOs and DONEs.
    'project': 'PROJECT',

    'resilient': False,
}

class TestOrg(unittest.TestCase):
    "Test org mode reading"

    def setUp(self):
        today = dt.datetime.now().replace(hour=random.randint(6, 10),
                                          minute=random.randint(0, 59))

        yesterday = today - dt.timedelta(days=1)
        tomorrow = today + dt.timedelta(days=1)
        accurate = today + dt.timedelta(hours=3)

        context = {
            'today': today.strftime(DAY),
            'yesterday': yesterday.strftime(DAY),
            'tomorrow': tomorrow.strftime(DAY),
            'accurate': accurate.strftime(DAYTIME),
        }

        tmpl = jinja2.Template(ORG_TMPL)

        self.rendered_org = tmpl.render(context)
        self.org_file = io.StringIO(self.rendered_org)
        self.db = orgnode.makelist(self.org_file,
                                   todo_default=['TODO', 'DONE', 'PROJECT'])


    def test_orgnode(self):
        "Test reading ORG using orgnode"
        self.assertIn('OPEN_TASK', self.db[1].tags)
        self.assertEqual(self.db[1].headline, 'This is open task')

    def test_conversion(self):
        "Test orgnode to events conversion"
        events = [
            helpers.orgnode_to_event(node, ORG_CONFIG)
            for node in self.db
        ]

        # If something fails intermittently - show debug data
        print(self.rendered_org)

        # Validate assumptions
        for event in events:
            print(event)

            if 'RANGE' in event.tags:
                self.assertEqual(event.priority, 'A')
                self.assertIn(DateType.RANGE, event.date_types)
                self.assertEqual(event.relevant_date.date_type,
                                 DateType.RANGE)
            if 'OPEN_TASK' in event.tags:
                self.assertEqual(event.priority, None)
                self.assertEqual(event.state.name, 'TODO')
                self.assertTrue(event.state.is_open)
                self.assertIn(DateType.SCHEDULED, event.date_types)
                self.assertEqual(event.relevant_date.date_type,
                                 DateType.SCHEDULED)
            if 'DEADLINE' in event.tags:
                relevant = event.relevant_date
                self.assertEqual(event.state.name, 'DONE')
                self.assertFalse(event.state.is_open)
                self.assertEqual(relevant.date_type,
                                 DateType.DEADLINE)
                self.assertFalse(relevant.appointment)
            if 'APP' in event.tags:
                relevant = event.relevant_date

                self.assertEqual(event.priority, 'B')
                self.assertTrue(relevant.appointment)

        self.assertEqual(len(events), 8)





import os
import yaml

"""
Handle config files with beautiful API
"""

class ConfigError(Exception):
    "Error while parsing config - show message and quit"

class Config:
    """
    Nice API for a configuration
    """

    def __init__(self, data, partial=None, path=None):
        "Create config out of dictionary"
        self._data = data
        self._partial = partial
        self._config_path = path

    @staticmethod
    def from_file(path):
        "Create config by reading a file"
        with open(path) as f:
            data = yaml.load(f)
        return Config(data, path=path)

    def get(self, key, default=None, required=True,
            assert_type=None, wrap=True):
        """
        Get a key from config.

        - Key can use dots to navigate nested config parts.
        - If key doesn't exist and is not required - the default value is returned
        - If key is required - ConfigError is raised.
        - Changing default to non-None assumes value is not required.
        - If assert_type is not None, the value type is asserted checked.
        - If wrap is True - treat dicts as sub-configs
        """
        key_split = key.split('.')

        # Find the deepest dictionary with the required value
        current = self._data
        for key_part in key_split[:-1]:
            if key_part in current:
                current = current[key_part]
            else:
                if required is True and default is None:
                    msg = "Required config key '%s' not found at level '%s'"
                    raise ConfigError(msg % (self._key_desc(key), key_part))
                else:
                    return default

        if key_split[-1] not in current:
            if required is True and default is None:
                msg = "Required config key '%s' not found"
                raise ConfigError(msg % self._key_desc(key))
            else:
                return default

        value = current[key_split[-1]]
        if assert_type is not None and not isinstance(value, assert_type):
            msg = "Key '%s' has type %s, but should have type %s"
            raise ConfigError(msg % (self._key_desc(key), type(value), assert_type))

        if wrap is True:
            # Wrap dictionaries
            if isinstance(value, dict):
                return Config(value, partial=key, path=self._config_path)

            # Wrap lists of dictionaries
            elif isinstance(value, list):
                if not value:
                    return value
                if set(type(el) for el in value) == {dict}:
                    return [
                        Config(element, partial='%s[%d]' % (key, i), path=self._config_path)
                        for i, element in enumerate(value)
                    ]
        return value

    def interpret_path(self, path):
        """
        Paths are relative to the config file and can use ~ to refer to HOME
        """
        path = os.path.expanduser(path)
        path = os.path.join(self._config_path, path)
        return path

    def get_path(self, key, default=None, required=True):
        "Get a path from config. Expand user ~, $HOME variables"
        value = self.get(key, default=default,
                         required=required, assert_type=str)
        if value is not None:
            value = self.interpret_path(value)
        return value

    def items(self, wrap=True):
        "Dict-like interface with wrapping"
        for key, value in self._data.items():
            if isinstance(value, dict) and wrap:
                yield (key, Config(value, partial=key, path=self._config_path))
            else:
                yield (key, value)

    def __iter__(self):
        "Allow casting config to dictionary"
        return self._data.items().__iter__()

    def __bool__(self):
        "For checking if config is empty"
        return bool(self._data)

    def _key_desc(self, key):
        "Describe key correctly, even if config is partial"
        if self._partial is None:
            return key
        return "(%s).%s" % (self._partial, key)

    def __getattr__(self, key):
        "Proxy access to valid fields of config"
        return self.get(key, required=True)

    def __repr__(self):
        return '<Config partial=%s %r>' % (self._partial, self._data)



"""
Org-mode compatible calendar implementation - handles a number of events in
time.
"""
import datetime as dt
import jinja2

from orgassist import log
from orgassist.calendar import DateType

class Calendar:
    """
    Manages multiple events, generates agenda
    """
    def __init__(self, agenda_path=None, agenda_content=None):
        "Initialize calendar"

        # Events sorted by sort_date
        self.events = []

        # Path to agenda
        self.agenda_path = agenda_path
        self.agenda_content = agenda_content
        assert self.agenda_content != self.agenda_path

    def add_events(self, events, internal_tag=None):
        "Add new events to the calendar"
        for event in events:
            event.meta['calendar_tag'] = internal_tag
        self.events += events
        self.events.sort()

    def del_events(self, internal_tag=None):
        "Delete events by internal tag"
        if internal_tag is None:
            self.events = []

        else:
            self.events = [
                event
                for event in self.events
                if event.meta['calendar_tag'] != internal_tag
            ]

    def get_unfinished(self, horizon,
                       list_unfinished_appointments,
                       relative_to=None):
        """
        Get a list of past unfinished events

        Args:
          horizon (datetime): The oldest date to consider.
          relative_to (datetime): The relative "now" time.
          list_unfinished_appointments (bool): Return all open or just scheduled.
        """
        if relative_to is None:
            relative_to = dt.datetime.now()

        print("GET UNFINISHED")
        unfinished = []
        for event in self.events:
            print("  ", event)
            date = event.relevant_date.sort_date
            if date < horizon:
                print("  BEFORE HORIZON")
                continue
            if event.state is None:
                print("  NO STATE, SO NOT AN OPEN STATE")
                continue
            if not event.state.is_open:
                print("  NOT OPEN STATE")
                continue
            if list_unfinished_appointments is False:
                print("  CHECK IF APP", event.date_types)
                if (DateType.SCHEDULED not in event.date_types and
                    DateType.DEADLINE not in event.date_types):
                    print("  NOT SCHEDULED")
                    continue
            if date > relative_to:
                # We are in future - not unfinished.
                break

            print("  GOT YOU")
            unfinished.append(event)
        return unfinished

    def get_appointments(self, since, horizon):
        "Get a list of scheduled and planned events"
        appointments = []

        print("GET APPOINTMENTS")
        appointments = []
        for event in self.events:
            # Include only appointments
            if not event.relevant_date.appointment:
                continue

            date = event.relevant_date.sort_date
            print("  ", event)
            if date < since:
                print("    IN PAST")
                continue
            if date > horizon:
                print("    OVER HORIZON")
                break

            print("  GOT YOU")
            appointments.append(event)
        return appointments

    def get_scheduled(self, horizon, relative_to=None):
        "Get tasks scheduled or deadlining in given period"

        print("GET SCHEDULED")
        scheduled = []
        for event in self.events:
            date = event.relevant_date.sort_date
            print("  ", event)
            if date < relative_to:
                print("    IN PAST")
                continue
            if date > horizon:
                print("    OVER HORIZON")
                break

            # State doesn't matter as long as the date is accurate
            if not event.relevant_date.appointment:
                continue

            print("  GOT YOU")
            scheduled.append(event)

        return scheduled

    def get_agenda(self, horizon_incoming, horizon_unfinished,
                   list_unfinished_appointments, relative_to=None):
        "Generate agenda in a text format"
        if relative_to is None:
            relative_to = dt.datetime.now()

        log.info("Getting agenda from %r to %r",
                 horizon_unfinished, horizon_incoming)

        # Open and read when needed so the file can be updated
        # without restarting bot.
        if self.agenda_content is not None:
            assert self.agenda_path is None
            content = self.agenda_content
        else:
            assert self.agenda_path is not None
            with open(self.agenda_path, 'r') as handle:
                content = handle.read()
        template = jinja2.Template(content)

        since = relative_to.replace(hour=0, minute=0)
        if (relative_to - since).total_seconds() < 4*60*60:
            # Include more past
            since -= dt.timedelta(hours=4)

        ctx = {
            'unfinished': self.get_unfinished(horizon_unfinished,
                                              list_unfinished_appointments,
                                              relative_to),
            # Incoming - from the whole day
            'appointments': self.get_appointments(since=since,
                                                  horizon=horizon_incoming),
            'now': relative_to,
        }

        return template.render(ctx)

    def __repr__(self):
        txt = "<Calendar events=%d>"
        return txt % len(self.events)




class Message:
    "API to unify all message data in one object"

    def __init__(self, text, sender, respond):
        self.text = text
        self.sender = sender
        self._respond = respond

    def respond(self, text):
        "Proxy to respond"
        self._respond(text)




from orgassist.config import ConfigError
from orgassist import log, templates

"""
Assistant class and assistant plugin interfaceo
"""

class PluginError(Exception):
    "Raised when plugin causes an error"

class Assistant:
    """
    Assistant keeps a state of a single communication.
    Identifies his boss on the bots interfaces (xmpp, irc, etc.)
    Dispatches incoming messages to commands.
    """
    # {'org': OrgContext, 'calendar': CalendarNotifications,
    #  'plugin_name': PluginClass }
    registered_plugins = {}

    def __init__(self, name, config, scheduler):
        "Initialize structures, plugins and validate configs early"
        # Assistant initialization
        self.assistant_name = name
        self.scheduler = scheduler
        self.config = config

        # Instances of plugins
        self.plugins = {}

        # Global assistant state to let plugins cooperate
        self.state = {}

        # Commands registered by plugins for dispatching
        # {command1: callback1, command2: callback1,
        #  command3: callback2 }
        self.commands = {}

        # List of callbacks to call boss when initiating communication
        self.boss_channels = []

        self._validate_config()
        self._initialize_plugins()

    def _initialize_plugins(self):
        "Create instances of plugins"
        # {name: handler1, name2: handler1, name3: handler2, ...}
        plugins = self.config.get('plugins', assert_type=dict)

        for plugin_name, plugin_config in plugins.items():
            plugin_cls = Assistant.registered_plugins.get(plugin_name, None)
            if plugin_cls is None:
                raise ConfigError("Configured plugin '%s' is not registered" %
                                  plugin_name)

            plugin = plugin_cls(self, plugin_config,
                                self.scheduler, self.state)
            plugin.validate_config()
            plugin.register()
            self.plugins[plugin_name] = plugin
            log.info('Plugin %s instantiated', plugin_name)

        # After all plugins are created - initialize plugins
        for plugin in self.plugins.values():
            plugin.initialize()


    def _validate_config(self):
        "Simple config validation - fail early"
        self.config.get('plugins')
        self.config.get('channels')

    def register_xmpp_bot(self, bot):
        """
        Dispatch to this assistant when a JID talks to bot with given
        resource.
        """
        for channel_cfg in self.config.channels:
            jid = channel_cfg.get('jid',
                                  required=False, assert_type=str)
            resource = channel_cfg.get('resource',
                                       required=False, assert_type=str)
            # FUTURE: Calling by name.
            # name = channel_cfg.get('name', default='boss')

            if not jid:
                continue

            # Incoming channel
            bot.add_dispatch(jid, resource,
                             self.handle_message)

            # Outgoing channel
            def create_closure(jid, resource):
                "Create closure containing JID and resource"
                def out(msg):
                    "Outgoing channel to the boss"
                    send_to = jid
                    if resource is not None:
                        send_to += '/' + resource
                    log.debug("Message to %s, body: %s", send_to, msg)
                    bot.send_message(send_to, msg)
                self.boss_channels.append(out)

            create_closure(jid, resource)

    def register_irc_bot(self, bot):
        "Register dispatch in an IRC bot"
        raise NotImplementedError

    def register_command(self, names, callback):
        "Register command dispatch"
        # TODO: Handle regular expressions as names
        if isinstance(names, str):
            names = [names]
        for name in names:
            name = name.lower().strip()
            if name in self.commands:
                raise PluginError("Command '%s' was already registered." % name)
            if ' ' in name:
                raise PluginError("Commands ('%s') can't have spaces within." % name)

            self.commands[name] = callback

    def handle_message(self, message):
        """
        Handle command sent by the Boss.
        """
        command = message.text.split()[0]
        command = command.lower().strip()
        handler = self.commands.get(command, None)
        if handler is not None:
            handler(message)
        else:
            message.respond(templates.get('DONT_UNDERSTAND'))

    def tell_boss(self, message):
        """
        Send message to boss using all registered channels.

        TODO: Allow to specifying priority or best channel.
        """
        for channel in self.boss_channels:
            channel(message)

    # Decorator to register context plugins
    @classmethod
    def plugin(cls, name):
        "Decorator to register plugins within assistant"
        def decorator(plugin_cls):
            "Register and return a plugin"
            if name in cls.registered_plugins:
                raise Exception("Plugin %s already defined" % name)
            cls.registered_plugins[name] = plugin_cls
            return plugin_cls
        return decorator


class AssistantPlugin:
    """
    Handles some data state (eg. org-mode directory),
    configures scheduler and may initiate communication.

    for_all[validate_config -> register] -> for_all[initialize]
    """

    def __init__(self, assistant, config, scheduler, state):
        """
        Args:
          assistant: Connected assistant object
          config: Part of config which is relevant to the plugin
          scheduler: Common scheduler which gets executed in the main loop
          state: a state shared between the plugins.
        """
        self.config = config
        self.scheduler = scheduler
        self.assistant = assistant
        self.state = state

    def register(self):
        """
        Register commands and other callbacks.

        Called first, before all plugins are created.
        """
        raise NotImplementedError

    def initialize(self):
        """
        Called once at the beginning to initialize - so implementors can leave
        __init__ alone. All plugins are registered when this method is called.
        """
        raise NotImplementedError

    def validate_config(self):
        """
        Validate configuration, raise ConfigError on problems.

        Touch all valid config options here, so that Config class can report
        what config keys were ignored (and are, for example, mistyped).
        """



"""
Helper functions to orgnode to handle incoming events and appointments in
org-mode.

Originally: Parse org-mode agendas, task lists, etc. and return simple
reminders to be included in environment status bar or shell.
"""

import re
import datetime as dt
import traceback as tb
import os
from collections import defaultdict

from orgassist import log
from orgassist.calendar import Event, EventState
from orgassist.calendar import EventDate, DateType

from . import orgnode

def unindent(body):
    "Unindent a 'common indent' of body of text"
    lines = body.split('\n')

    if not lines:
        return body

    # Find possible indent in the first line
    m = re.match('^([ \t]+)', lines[0])
    if not m:
        # No ident found
        return body
    indent = m.group()

    # Find a smallest common indent in all lines (spaces/tabs mixed)
    for line in lines[1:]:
        if not line.strip():
            # Ignore lines without a text
            continue

        for i, char in enumerate(indent):
            if len(line) == i or char != line[i]:
                # Cut indent short
                indent = indent[:i]
                break

    # Strip indent
    unindented = [
        line.replace(indent, '', 1).rstrip()
        for line in lines
    ]

    # Strip empty lines from beginning and end.
    while unindented and not unindented[0].strip():
        unindented = unindented[1:]

    while unindented and not unindented[-1].strip():
        unindented = unindented[:-1]

    return "\n".join(unindented)


def load_orgnode(cfg):
    "Load data from all org-files using orgnode"
    # Aggregated Orgnodes objects
    db = []

    todo_all = set(cfg['todos_open'])
    todo_all.update(cfg['todos_closed'])
    todo_all.add(cfg['project'])

    for path in cfg['files']:
        db += orgnode.makelist(path, todo_default=todo_all)

    if not cfg['files_re']:
        return db

    first = True
    # Read by regexp
    regexp = re.compile(cfg['files_re'], flags=re.UNICODE)
    for root, dirs, files in os.walk(cfg['base'],
                                     followlinks=True):
        if '.git' in dirs:
            dirs.remove('.git')

        for filename in files:
            if not re.match(regexp, filename):
                continue

            path = os.path.join(root, filename)
            try:
                db += orgnode.makelist(path,
                                       todo_default=todo_all)
            except Exception:
                if cfg['resilient']:
                    log.warning("Warning: Ignoring error while parsing %s", path)
                    if first:
                        tb.print_exc()
                    first = False
                    continue
                raise

    return db

def orgnode_to_event(node, org_config, relative_to=None):
    "Convert orgnode entries to events"
    event = Event(node.headline)
    event.add_tags(node.tags)

    event.body = unindent(node.body)
    if node.todo:
        state = EventState(node.todo,
                           is_open=node.todo in org_config['todos_open'])
        event.set_state(state)
    if node.priority:
        event.priority = node.priority

    if node.scheduled:
        date = EventDate(node.scheduled, DateType.SCHEDULED)
        event.add_date(date, relative_to)

    if node.deadline:
        date = EventDate(node.deadline, DateType.DEADLINE)
        event.add_date(date, relative_to)

    for ranged in node.rangelist:
        date = EventDate(ranged, DateType.RANGE)
        event.add_date(date, relative_to)

    for node_date in node.datelist:
        # Appointment dates
        date = EventDate(node_date, DateType.TIMESTAMP)
        event.add_date(date, relative_to)

    return event


"""
def until(date, relative):
    "
    Return time difference taking into account that date might not have a time given.
    If so - assume end of day. `relative' always has a time.
    "

    # Decorate with time
    if not isinstance(date, dt.datetime):
        date = dt.datetime(date.year, date.month, date.day, 23, 59, 59)
        appointment = False
    else:
        # Appointment is an event with accurate time defined.
        appointment = True

    delta = (date - relative).total_seconds()
    # If negative - then a past event.
    return date, delta / 60.0 / 60.0 / 24.0, appointment


def closest(date_list, relative):
    "Find closest future event relative to a given date"
    # Closest date - original
    closest_date = None

    # Closest date converted to datetime
    closest_converted_date = None

    # Time delta for closest date
    closest_delta = None

    # Is the date an appointment? (accurate to hour:minute?)
    appointment = None

    for date in sorted(date_list):
        converted_date, days, appointment = until(date, relative)

        closest_date = date
        closest_converted_date = converted_date
        closest_delta = days

        if closest_delta < 0:
            # Past event, iterate more
            continue
        # This is a first future event, do not check more.
        break

    return {
        'converted_date': closest_converted_date,
        'date': closest_date,
        'delta': closest_delta,
        'appointment': appointment
    }

def get_totals_stat(db, cfg):
    "
    Count all entries versus not ignored (opened).

    This is supposed to help push TODOs without a specified execution time.
    "
    count_total = 0
    count_open = 0

    for entry in db:
        # Iterate over entries

        if not entry.todo:
            # No designation at all, not a `task'
            continue

        if (entry.datelist or entry.scheduled or
            entry.deadline or entry.rangelist):
            # Has time - is already counted elsewhere
            continue

        count_total += 1

        # Ignore ones marked as "done/finished/closed"
        if entry.todo not in cfg['todos_closed']:
            count_open += 1

    return count_open, count_total

"""

def get_incoming(db, cfg):
    """
    Parse all events and gather ones happening in the near
    future (set by horizont) and unfinished from past.

    Takes into account any possible dates, can report the same
    event multiple times for different dates.
    """
    today = dt.datetime.today()

    # Returned aggregation
    ret = {
        # Incoming events
        'incoming': [],

        # Past TODO events (SCHEDULED, DEADLINE) not marked as DONE
        'unfinished': [],

        # Things to show, to remind you, you're responsible
        # project entry -> {stats}
        'projects': {},

        # Found, but over-horizon
        'filtered_future': 0,
        'filtered_past': 0,
    }

    for entry in db:
        # Iterate over entries
        if entry.todo and entry.parent in ret['projects']:
            # Count number of tasks open within a project.
            # DONE, TODO, all types
            current_entry = entry
            while current_entry.parent:
                current_entry = current_entry.parent
                if current_entry in ret['projects']:
                    ret['projects'][current_entry][entry.todo] += 1

        # Now, ignore ones marked as "done/finished/closed"
        if entry.todo in cfg['todos_closed']:
            continue

        def analyze_dates(event, dates, datetype):
            data = closest(dates, relative=today)

            data.update({
                'eventtype': datetype,
                'headline': entry.headline,
                'body': unindent(entry.body),

                'entry': entry,
            })

            if data['delta'] is None:
                return # No dates, no event.
            elif data['delta'] < 0:
                # Past event
                if (cfg['horizont_past'] is not None and
                    data['delta'] >= -cfg['horizont_past']):

                    ret['unfinished'].append((data['converted_date'], data))
                else:
                    ret['filtered_past'] += 1
            else:
                # Future event
                if data['delta'] <= cfg['horizont_future']:
                    ret['incoming'].append((data['converted_date'], data))
                else:
                    ret['filtered_future'] += 1

        event = calendar.Event(entry.headline)
        event.body = unindent(entry.body)
        event.add_tags(entry.tags)

        analyze_dates(event, entry.datelist, 'TIMESTAMP')

        if entry.rangelist:
            starts = [dr[0] for dr in entry.rangelist]
            analyze_dates(event, starts, 'RANGE')

        scheduled = entry.scheduled
        deadline = entry.deadline

        if scheduled:
            analyze_dates(event, [scheduled], "SCHEDULED")

        if deadline:
            analyze_dates(event, [deadline], "DEADLINE")

        if entry.todo == cfg['project']:
            ret['projects'][entry] = defaultdict(lambda: 0)

    return ret



"""
The main plugin - org-mode plugin.

- Sent notifications before events
- Remind automatically about your todays agenda.
- Allow note-taking on the go.

Currently uses orgnode, later it might be sensible to use something newer.
Works for me though.

(C) 2018 by Tomasz bla Fortuna
"""
from orgassist import log
from orgassist.assistant import Assistant, AssistantPlugin

from . import helpers


@Assistant.plugin('org')
class OrgPlugin(AssistantPlugin):
    """
    Handle operations on an org-mode tree
    """
    def refresh_db(self):
        "Refresh/load DB with org entries"
        log.info('Refreshed/read org-mode data')
        db = helpers.load_orgnode(self.parsed_config)
        events = [
            helpers.orgnode_to_event(node, self.parsed_config)
            for node in db
        ]

        # TODO: Handle TODOs too, not only the appointments/schedules
        events = [
            event
            for event in events
            if event.relevant_date is not None
        ]

        self.state['calendar'].del_events('org')
        self.state['calendar'].add_events(events, 'org')
        return events

    def register(self):
        commands = [
            (['org:note', 'note', 'no'], self.handle_note),
            (['org:refresh'], self.handle_refresh),
        ]
        for aliases, callback in commands:
            self.assistant.register_command(aliases, callback)

    def initialize(self):
        "Initialize org plugin, read database and schedule updates"
        self.refresh_db()

        interval = self.config.get('scan_interval_s', assert_type=int)
        self.scheduler.every(interval).seconds.do(self.refresh_db)

        #self.scheduler.every(10).seconds.do(lambda:
        # self.assistant.tell_boss('You are a bad person.\nTesttest\ntesttest'))

    def validate_config(self):
        "Read config and apply defaults"
        self.parsed_config = {
            # Include those files (full path)
            'files': self.config.get('files', default=[]),

            # Scan given base for all files matching regexp
            'files_re': self.config.get('org_regexp', default=r'.*\.org$'),
            'base': self.config.get_path('directory'),

            # Look 5 days ahead
            #'horizont_future': self.config.get('agenda.horizont_future', default=2),
            #'horizont_past': self.config.get('agenda.horizont_past', default=10),

            'todos_open': self.config.get('todos.open', default=['TODO']),
            'todos_closed': self.config.get('todos.closed', default=['DONE', 'CANCELLED']),

            # How grouping entry is marked - which groups TODOs and DONEs.
            'project': self.config.get('todos.project', default='PROJECT'),

            # Ignore exceptions during file parsing (happens in orgnode when file is
            # badly broken or not ORG at all).
            # This should be fixed in orgnode.
            'resilient': False,
        }

    def handle_note(self, message):
        "Take a note"
        pass

    def handle_refresh(self, message):
        "Handle refresh request"
        events = self.refresh_db()
        message.respond("Loaded %d events" % len(events))
